from datetime import datetime
from pathlib import Path
from unittest.mock import patch

import numpy as np
import pytest
import zmq
from iCalibrationDB import Conditions, ConstantMetaData, Constants

from cal_tools.agipdlib import AgipdCorrections, CellRange
from cal_tools.plotting import show_processed_modules
from cal_tools.tools import (
    get_dir_creation_date,
    get_from_db,
    get_pdu_from_db,
    module_index_to_qm,
    send_to_db,
)

# AGIPD operating conditions.
ACQ_RATE = 1.1
BIAS_VOLTAGE = 300
GAIN_SETTING = 0
INTEGRATION_TIME = 12
MEM_CELLS = 352
PHOTON_ENERGY = 9.2

AGIPD_KARABO_ID = "SPB_DET_AGIPD1M-1"
WRONG_AGIPD_MODULE = "AGIPD_**"

CAL_DB_INTERFACE = "tcp://max-exfl017:8020"
WRONG_CAL_DB_INTERFACE = "tcp://max-exfl017:0000"


@pytest.fixture
def _agipd_const_cond():
    # AGIPD dark offset metadata
    constant = Constants.AGIPD.Offset()

    condition = Conditions.Dark.AGIPD(
        memory_cells=MEM_CELLS,
        bias_voltage=BIAS_VOLTAGE,
        acquisition_rate=ACQ_RATE,
        gain_setting=GAIN_SETTING,
    )
    return constant, condition


def test_show_processed_modules():
    mnames = ['Q1M1']

    with pytest.raises(ValueError) as err:
        show_processed_modules('LDP', mnames=mnames,
                               constants={}, mode='processed')
        assert 'LDP' in err.value()


@pytest.mark.requires_gpfs
def test_dir_creation_date():
    folder = '/gpfs/exfel/exp/CALLAB/202031/p900113/raw'

    date = get_dir_creation_date(folder, 9983)
    assert isinstance(date, datetime)
    assert str(date) == '2020-09-23 13:30:50+00:00'

    with pytest.raises(ValueError) as e:
        get_dir_creation_date(folder, 4)
    assert e.value.args[1] == Path(folder) / 'r0004'

    # The following data predates the addition of creation_time in metadata
    date = get_dir_creation_date(folder, 9999)
    assert isinstance(date, datetime)
    assert str(date) == '2019-12-16 08:52:25.196603'


def _call_get_from_db(
    constant,
    condition,
    karabo_id,
    karabo_da,
    load_data=True,
    cal_db_interface=CAL_DB_INTERFACE,
    creation_time=None,
    doraise=True,
    timeout=10000,
):

    data, metadata = get_from_db(
        karabo_id=karabo_id,
        karabo_da=karabo_da,
        constant=constant,
        condition=condition,
        empty_constant=None,
        cal_db_interface=cal_db_interface,
        creation_time=creation_time,
        meta_only=True,
        load_data=load_data,
        ntries=1,
        doraise=doraise,
        timeout=timeout,
    )
    return data, metadata


def _call_send_to_db(
    constant,
    condition,
    db_module,
    data=np.zeros((2, 2, 2)),
    cal_db_interface=CAL_DB_INTERFACE,
    report_path="",
    doraise=True,
    timeout=1000,
    ntries=1,
):

    # TODO: create a known_constant for testing.
    constant.data = data
    metadata = send_to_db(
        karabo_id=AGIPD_KARABO_ID,
        db_module=db_module,
        constant=constant,
        condition=condition,
        file_loc="proposal, runs",
        report_path=report_path,
        cal_db_interface=cal_db_interface,
        creation_time=None,
        ntries=ntries,
        doraise=doraise,
        timeout=timeout,
    )
    return metadata


# TODO add a marker for accessing zmq end_point
@pytest.mark.requires_gpfs
def test_get_from_db_load_data(_agipd_const_cond):
    """ Test retrieving calibration constants with get_from_db
    with different loading data scenarios.
    """
    constant, condition = _agipd_const_cond

    # Normal operation and loading data from h5file.
    data, md = _call_get_from_db(
        constant=constant, condition=condition,
        karabo_id=AGIPD_KARABO_ID, karabo_da="AGIPD00",
    )

    assert type(data) is np.ndarray
    assert isinstance(md, ConstantMetaData)

    # None karabo_id is given.
    data, md = _call_get_from_db(
        constant=constant, condition=condition,
        karabo_id=None, karabo_da="AGIPD00",
    )

    assert data is None
    assert md is None

    # Retrieve constant without loading the data.
    data, md = _call_get_from_db(
        constant=constant, condition=condition,
        karabo_id=AGIPD_KARABO_ID, karabo_da="AGIPD00",
        load_data=False,
    )

    assert data is None
    assert isinstance(md, ConstantMetaData)


# TODO add a marker for accessing zmq end_point
@pytest.mark.requires_gpfs
def test_raise_get_from_db(_agipd_const_cond):
    """ Test error raised scenarios for get_from_db:"""

    constant, condition = _agipd_const_cond

    # Wrong address for the calibration database.
    with pytest.raises(zmq.error.Again) as excinfo:
        _call_get_from_db(
            constant=constant, condition=condition,
            karabo_id=AGIPD_KARABO_ID, karabo_da="AGIPD00",
            cal_db_interface=WRONG_CAL_DB_INTERFACE,
        )
    assert str(excinfo.value) == "Resource temporarily unavailable"

    # Wrong type for creation_time.
    with pytest.raises(ValueError):
        _call_get_from_db(
            constant=constant, condition=condition,
            karabo_id=AGIPD_KARABO_ID, karabo_da="AGIPD00",
            creation_time="WRONG_CREATION_TIME",
        )

    # No constant file path metadata retrieved.
    with patch("iCalibrationDB.ConstantMetaData.retrieve", return_value=""):
        with pytest.raises(ValueError):
            _call_get_from_db(
                constant=constant, condition=condition,
                karabo_id=AGIPD_KARABO_ID, karabo_da="AGIPD00",
            )


def test_no_doraise_get_from_db(_agipd_const_cond):
    """get_from_db using wrong cal_db_interface
    fails without raising errors, as doraise = False
    """
    constant, condition = _agipd_const_cond

    data, _ = _call_get_from_db(
        constant=constant, condition=condition,
        karabo_id=AGIPD_KARABO_ID, karabo_da="AGIPD00",
        cal_db_interface=WRONG_CAL_DB_INTERFACE,
        doraise=False,
    )
    assert data is None


@patch(
    'iCalibrationDB.ConstantMetaData.send',
    return_value='',
)
def test_send_to_db_success(send, _agipd_const_cond):
    """test sending constants to the database (send_to_db):
    Injecting constant as expected.
    # TODO: Add a test calibration constant to the test physical module
    # to inject without mocking `send` method.
    """
    # Use wrong AGIPD module as a backup.
    # To avoid constants injection, in-case of mock failure.

    constant, condition = _agipd_const_cond

    db_module = WRONG_AGIPD_MODULE
    metadata = _call_send_to_db(
        constant=constant,
        condition=condition,
        db_module=db_module,
    )

    assert isinstance(metadata, ConstantMetaData)


@patch(
    'iCalibrationDB.ConstantMetaData.send',
    return_value='',
)
def test_raise_send_to_db_mocked(send, _agipd_const_cond):
    """Test raised errors while sending constants to the
    database (send_to_db):
    """
    # Use wrong AGIPD module as a backup.
    # To avoid constants injection, in-case of mock failure.
    constant, condition = _agipd_const_cond

    # report_path has the wrong type.
    with pytest.raises(TypeError):
        _call_send_to_db(
            constant=constant,
            condition=condition,
            db_module=WRONG_AGIPD_MODULE,
            report_path=2,
        )

    # No constant data to inject.
    with pytest.raises(ValueError):
        _call_send_to_db(
            constant=constant,
            condition=condition,
            db_module=WRONG_AGIPD_MODULE,
            data=None,
        )


def test_raise_send_to_db(_agipd_const_cond):

    constant, condition = _agipd_const_cond

    # wrong calibration database address.
    with pytest.raises(zmq.error.Again) as excinfo:
        _call_send_to_db(
            constant=constant,
            condition=condition,
            db_module=WRONG_AGIPD_MODULE,
            cal_db_interface=WRONG_CAL_DB_INTERFACE,
        )
    assert str(excinfo.value) == "Resource temporarily unavailable"


@pytest.mark.requires_caldb
def test_get_pdu_from_db(_agipd_const_cond):

    constant, condition = _agipd_const_cond
    snapshot_at = "2021-05-06 00:20:10.00"

    # A karabo_da str returns a list of one element.
    pdu_dict = get_pdu_from_db(karabo_id="TEST_DET_CAL_CI-1",
                               karabo_da="TEST_DET_CAL_DA0",
                               constant=constant,
                               condition=condition,
                               cal_db_interface=CAL_DB_INTERFACE,
                               snapshot_at=snapshot_at,
                               timeout=30000)
    assert len(pdu_dict) == 1
    assert pdu_dict[0] == "CAL_PHYSICAL_DETECTOR_UNIT-1_TEST"

    # A list of karabo_das to return thier PDUs, if available.
    pdu_dict = get_pdu_from_db(karabo_id="TEST_DET_CAL_CI-1",
                               karabo_da=[
                                   "TEST_DET_CAL_DA0",
                                   "TEST_DET_CAL_DA1",
                                   "UNAVAILABLE_DA"],
                               constant=constant,
                               condition=condition,
                               cal_db_interface=CAL_DB_INTERFACE,
                               snapshot_at=snapshot_at,
                               timeout=30000)

    assert pdu_dict == ["CAL_PHYSICAL_DETECTOR_UNIT-1_TEST",
                        "CAL_PHYSICAL_DETECTOR_UNIT-2_TEST",
                        None]

    # "all" is used to return all corresponding units for a karabo_id.
    pdu_dict = get_pdu_from_db(karabo_id="TEST_DET_CAL_CI-1",
                               karabo_da="all",
                               constant=constant,
                               condition=condition,
                               cal_db_interface=CAL_DB_INTERFACE,
                               snapshot_at=snapshot_at,
                               timeout=30000)

    assert len(pdu_dict) == 2
    assert pdu_dict == ["CAL_PHYSICAL_DETECTOR_UNIT-1_TEST",
                        "CAL_PHYSICAL_DETECTOR_UNIT-2_TEST"]


# TODO add a marker for accessing zmq end_point
@pytest.mark.requires_gpfs
@pytest.mark.requires_caldb
def test_initialize_from_db():
    creation_time = datetime.strptime(
        "2020-01-07 13:26:48.00", "%Y-%m-%d %H:%M:%S.%f")

    agipd_corr = AgipdCorrections(
        max_cells=MEM_CELLS,
        cell_sel=CellRange([0, 500, 1], MEM_CELLS))

    agipd_corr.allocate_constants(
        modules=[0],
        constant_shape=(3, MEM_CELLS, 512, 128))

    dark_const_time_dict = agipd_corr.initialize_from_db(
        karabo_id="TEST_DET_CAL_CI-1",
        karabo_da="TEST_DET_CAL_DA1",
        cal_db_interface=CAL_DB_INTERFACE,
        creation_time=creation_time,
        memory_cells=MEM_CELLS,
        bias_voltage=BIAS_VOLTAGE,
        photon_energy=PHOTON_ENERGY,
        gain_setting=GAIN_SETTING,
        acquisition_rate=ACQ_RATE,
        integration_time=INTEGRATION_TIME,
        module_idx=0,
        only_dark=False,
    )

    assert dark_const_time_dict == {
        "Offset": None,
        "Noise": None,
        "ThresholdsDark": None,
        "BadPixelsDark": None,
    }

    dark_const_time_dict = agipd_corr.initialize_from_db(
        karabo_id=AGIPD_KARABO_ID,
        karabo_da="AGIPD00",
        cal_db_interface=CAL_DB_INTERFACE,
        creation_time=creation_time,
        memory_cells=MEM_CELLS, bias_voltage=BIAS_VOLTAGE,
        photon_energy=PHOTON_ENERGY, gain_setting=GAIN_SETTING,
        integration_time=INTEGRATION_TIME,
        acquisition_rate=ACQ_RATE, module_idx=0,
        only_dark=False,
    )

    # A retrieved constant has a value of datetime creation_time
    assert isinstance(dark_const_time_dict["Offset"], datetime)
    assert list(dark_const_time_dict.keys()) == [
        "Offset", "Noise", "ThresholdsDark", "BadPixelsDark"]


def test_module_index_to_qm():

    assert module_index_to_qm(0) == 'Q1M1'
    assert module_index_to_qm(1) == 'Q1M2'
    assert module_index_to_qm(4) == 'Q2M1'
    assert module_index_to_qm(6) == 'Q2M3'

    assert module_index_to_qm(4, 5) == 'Q5M1'

    with pytest.raises(AssertionError):
        module_index_to_qm(18)

    with pytest.raises(AssertionError):
        module_index_to_qm(7, 5)
